# i dont know the math i stole it from https://docs.pytorch.org/tutorials/intermediate/reinforcement_q_learning.html

import math
import random
from collections import namedtuple, deque
from itertools import count

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

device = torch.device(
    "cuda" if torch.cuda.is_available() else
    "mps" if torch.backends.mps.is_available() else
    "cpu"
)

Transition = namedtuple('Transition',
                        ('state', 'action', 'next_state', 'reward'))

class ReplayMemory(object):

    def __init__(self, capacity):
        self.memory = deque([], maxlen=capacity)

    def push(self, *args):
        self.memory.append(Transition(*args))

    def sample(self, batch_size):
        return random.sample(self.memory, batch_size)

    def __len__(self):
        return len(self.memory)


class DQN(nn.Module):

    def __init__(self, n_observations, n_actions, layers):
        super(DQN, self).__init__()
        self.layers = [];
        self.layers.append(nn.Linear(n_observations, layers[0]))
        for i, lay in enumerate(layers[:-1]):
            self.layers.append(nn.Linear(lay, layers[i+1]))
        self.layers.append(nn.Linear(layers[len(layers) - 1], n_actions))
        self.layers = nn.Sequential(*[item for it in [[l, nn.ReLU()] for l in self.layers] for item in it][:-1])

    def forward(self, x):
        return self.layers(x)

class Learner:

    def __init__(self,
        n_actions,
        n_observations,
        state = None,
        layers = [128, 128],
        BATCH_SIZE = 128,
        GAMMA = 0.99,
        EPS_START = 0.9,
        EPS_END = 0.05,
        EPS_DECAY = 1000,
        TAU = 0.005,
        LR = 1e-4
    ):
        self.BATCH_SIZE = BATCH_SIZE
        self.GAMMA = GAMMA
        self.EPS_START = EPS_START
        self.EPS_END = EPS_END
        self.EPS_DECAY = EPS_DECAY
        self.TAU = TAU
        self.LR = LR
        self.n_observations = n_observations
        self.n_actions = n_actions
        if state is not None and 'layers' in state:
            layers = state['layers'];
        self.layers = layers
        self.policy_net = DQN(n_observations, n_actions, layers).to(device)
        self.steps_done = 0
        if state is not None:
            self.policy_net.load_state_dict(state['model'])
            self.steps_done = state['steps_done']
        self.target_net = DQN(n_observations, n_actions, layers).to(device)
        self.target_net.load_state_dict(self.policy_net.state_dict())
        self.state = None

        self.optimizer = optim.AdamW(self.policy_net.parameters(), lr=self.LR, amsgrad=True)
        self.memory = ReplayMemory(10000)

    def get_model_state(self):
        return { 'model': self.policy_net.state_dict(), 'steps_done': self.steps_done, 'layers': self.layers }

    def select_action(self, state, final=False):
        sample = random.random()
        if final:
            eps_threshold = 0
        else:
            eps_threshold = self.EPS_END + (self.EPS_START - self.EPS_END) * \
                math.exp(-1. * self.steps_done / self.EPS_DECAY)
            self.steps_done += 1
        if sample > eps_threshold:
            with torch.no_grad():
                return self.policy_net(state).max(1).indices.view(1, 1)
        else:
            return torch.tensor([[random.randrange(0, self.n_actions)]], device=device, dtype=torch.long)

    def optimize_model(self):
        memory = self.memory
        optimizer = self.optimizer
        policy_net = self.policy_net

        if len(memory) < self.BATCH_SIZE:
            return

        transitions = memory.sample(self.BATCH_SIZE)
        batch = Transition(*zip(*transitions))
        non_final_mask = torch.tensor(tuple(map(lambda s: s is not None,
                                              batch.next_state)), device=device, dtype=torch.bool)
        non_final_next_states = torch.cat([s for s in batch.next_state
                                                    if s is not None])
        state_batch = torch.cat(batch.state)
        action_batch = torch.cat(batch.action)
        reward_batch = torch.cat(batch.reward)

        state_action_values = policy_net(state_batch).gather(1, action_batch)

        next_state_values = torch.zeros(self.BATCH_SIZE, device=device)
        with torch.no_grad():
            next_state_values[non_final_mask] = self.target_net(non_final_next_states).max(1).values
        expected_state_action_values = (next_state_values * self.GAMMA) + reward_batch

        criterion = nn.SmoothL1Loss()
        loss = criterion(state_action_values, expected_state_action_values.unsqueeze(1))

        optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_value_(policy_net.parameters(), 100)
        optimizer.step()

    def decide(self, state, final=False):
        state = torch.tensor(state, dtype=torch.float32, device=device).unsqueeze(0)
        action = self.select_action(state, final=False)
        return action

    def learn(self, state, action, observation, reward):
        state = torch.tensor(state, dtype=torch.float32, device=device).unsqueeze(0)

        if not observation:
            next_state = None
        else:
            next_state = torch.tensor(observation, dtype=torch.float32, device=device).unsqueeze(0)

        reward = torch.tensor([reward], device=device)

        self.memory.push(state, action, next_state, reward)

        self.optimize_model()

        target_net_state_dict = self.target_net.state_dict()
        policy_net_state_dict = self.policy_net.state_dict()
        for key in policy_net_state_dict:
            target_net_state_dict[key] = policy_net_state_dict[key]*self.TAU + target_net_state_dict[key]*(1-self.TAU)
        self.target_net.load_state_dict(target_net_state_dict)

